# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

'''apport package hook for update-manager

(c) 2011 Canonical Ltd.
Author: Brian Murray <brian@ubuntu.com>
'''

import os
import re
import subprocess
from apport.hookutils import (
    attach_gsettings_package, attach_root_command_outputs,
    attach_file_if_exists, command_available,
    recent_syslog)


def run_hwe_command(option):
    command = ['hwe-support-status', option]
    sp = subprocess.Popen(command, stdout=subprocess.PIPE,
                          stderr=subprocess.STDOUT,
                          stdin=None)
    out = sp.communicate()[0]
    if sp.returncode == 0:
        res = out.strip()
    # exit code is 10 on unsupported HWE
    elif sp.returncode == 10:
        res = out.strip()
    else:
        res = (b'Error: command ' + str(command).encode()
               + b' failed with exit code '
               + str(sp.returncode).encode() + b': ' + out)
    return res


def add_info(report, ui):

    problem_type = report.get("ProblemType", None)
    if problem_type == "Bug":
        response = ui.yesno("Is the issue you are reporting one you \
encountered when upgrading Ubuntu from one release to another?")
    else:
        response = None
    if response:
        os.execlp('apport-bug', 'apport-bug', 'ubuntu-release-upgrader')
    else:
        attach_gsettings_package(report, 'update-manager')
        attach_file_if_exists(report, '/var/log/apt/history.log',
                              'DpkgHistoryLog.txt')
        attach_file_if_exists(report, '/var/log/apt/term.log',
                              'DpkgTerminalLog.txt')
        attach_root_command_outputs(
            report,
            {'CurrentDmesg.txt':
                'dmesg | comm -13 --nocheck-order /var/log/dmesg -'})
        if command_available('hwe-support-status'):
            # not using apport's command_output because it doesn't expect a
            # return code of 10
            unsupported = run_hwe_command('--show-all-unsupported')
            if unsupported:
                report['HWEunsupported'] = unsupported
                report['HWEreplacements'] = \
                    run_hwe_command('--show-replacements')
        report["Aptdaemon"] = recent_syslog(re.compile("AptDaemon"))



#!/usr/bin/env python
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

"""Integration of package managers into UpdateManager"""
# (c) 2005-2009 Canonical, GPL

from __future__ import absolute_import

import gi
gi.require_version('Snapd', '1')
from gi.repository import GLib, Gtk, Snapd

from apt import Cache
import json
import logging
import os
import re
import subprocess
from gettext import gettext as _
from threading import Thread

from UpdateManager.Core.MyCache import MyCache
from UpdateManager.Core.utils import inhibit_sleep, get_dist_version
from UpdateManager.Dialogs import Dialog


class InstallBackend(Dialog):
    ACTION_UPDATE = 0
    ACTION_INSTALL = 1

    def __init__(self, window_main, action):
        Dialog.__init__(self, window_main)
        self.action = action
        self.sleep_cookie = None

    def start(self):
        os.environ["APT_LISTCHANGES_FRONTEND"] = "none"

        # Do not suspend during the update process
        self.sleep_cookie = inhibit_sleep()

        if self.action == self.ACTION_INSTALL:
            # Get the packages which should be installed and update
            pkgs_install = []
            pkgs_upgrade = []
            pkgs_remove = []
            # Get a fresh cache in case update-manager's is outdated to
            # skip operations that already took place
            fresh_cache = Cache(rootdir=self.window_main.cache.rootdir)
            for pkg in self.window_main.cache:
                try:
                    if pkg.marked_install \
                       and not fresh_cache[pkg.name].is_installed:
                        pkgname = pkg.name
                        if pkg.is_auto_installed:
                            pkgname += "#auto"
                        pkgs_install.append(pkgname)
                    elif (pkg.marked_upgrade
                          and fresh_cache[pkg.name].is_upgradable):
                        pkgs_upgrade.append(pkg.name)
                    elif (pkg.marked_delete
                          and fresh_cache[pkg.name].is_installed):
                        pkgs_remove.append(pkg.name)
                except KeyError:
                    # pkg missing from fresh_cache can't be modified
                    pass
            self.commit(pkgs_install, pkgs_upgrade, pkgs_remove)
        else:
            self.update()

    def update(self):
        """Run a update to refresh the package list"""
        raise NotImplementedError

    def commit(self, pkgs_install, pkgs_upgrade, pkgs_remove):
        """Commit the cache changes """
        raise NotImplementedError

    def get_snap_seeds(self):
        seeded_snaps = {}
        unseeded_snaps = {}

        curr_channel = "stable/ubuntu-" + get_dist_version()

        try:
            d2s_file = open(
                '/usr/share/ubuntu-release-upgrader/deb2snap.json', 'r')
            d2s = json.load(d2s_file)
            d2s_file.close()

            for snap in d2s["seeded"]:
                seed = d2s["seeded"][snap]
                deb = seed.get("deb", None)
                to_channel = seed.get("to_channel", curr_channel)
                seeded_snaps[snap] = (deb, to_channel)

            for snap in d2s["unseeded"]:
                unseed = d2s["unseeded"][snap]
                from_channel = unseed.get("from_channel", curr_channel)
                unseeded_snaps[snap] = (from_channel)
        except Exception as e:
            logging.debug("error reading deb2snap.json file (%s)" % e)

        return seeded_snaps, unseeded_snaps

    def get_deb2snap_dups(self):
        # update and grab the latest cache
        try:
            if self.window_main.cache is None:
                self.window_main.cache = MyCache(None)
            else:
                self.window_main.cache.open(None)
                self.window_main.cache._initDepCache()
            cache = self.window_main.cache
        except Exception as e:
            logging.debug("error reading cache (%s)" % e)
            cache = None

        duplicates = []
        seeded_snaps, _ = self.get_snap_seeds()

        for snap, (deb, _) in seeded_snaps.items():
            # if the deb is installed and was not manually installed,
            # replace it
            if (deb in cache and cache[deb].is_installed):
                deb_is_auto = True
                cache[deb].mark_delete()

                for pkg in cache.get_changes():
                    if (pkg.is_installed and pkg.marked_delete
                       and not pkg.is_auto_installed):
                        deb_is_auto = False
                        break

                cache.clear()

                if deb_is_auto:
                    duplicates.append(deb)

        return duplicates

    def get_snap_transitions(self):
        # populate snap_list with deb2snap transitions
        snap_list = {}
        seeded_snaps, unseeded_snaps = self.get_snap_seeds()

        for snap, (deb, to_channel) in seeded_snaps.items():
            snap_object = {}
            # check if the snap is already installed
            snap_info = subprocess.Popen(["snap", "info", snap],
                                         universal_newlines=True,
                                         stdout=subprocess.PIPE).communicate()
            if re.search("^installed: ", snap_info[0], re.MULTILINE):
                logging.debug("Snap %s is installed" % snap)
                continue
            elif (deb in self.window_main.duplicate_packages):
                # install the snap if the deb was just marked delete
                snap_object['command'] = 'install'
                snap_object['channel'] = to_channel
                snap_list[snap] = snap_object

        for snap, (from_channel) in unseeded_snaps.items():
            snap_object = {}
            # check if the snap is already installed
            snap_info = subprocess.Popen(["snap", "info", snap],
                                         universal_newlines=True,
                                         stdout=subprocess.PIPE).communicate()
            if re.search("^installed: ", snap_info[0], re.MULTILINE):
                logging.debug("Snap %s is installed" % snap)
                # its not tracking the release channel so don't remove
                re_channel = "stable/ubuntu-[0-9][0-9].[0-9][0-9]"
                if not re.search(r"^tracking:.*%s" % re_channel,
                                 snap_info[0], re.MULTILINE):
                    logging.debug("Snap %s is not tracking the release channel"
                                  % snap)
                    continue

                snap_object['command'] = 'remove'

                # check if this snap is being used by any other snaps
                conns = subprocess.Popen(["snap", "connections", snap],
                                         universal_newlines=True,
                                         stdout=subprocess.PIPE).communicate()

                for conn in conns[0].split('\n'):
                    conn_cols = conn.split()
                    if len(conn_cols) != 4:
                        continue
                    plug = conn_cols[1]
                    slot = conn_cols[2]

                    if slot.startswith(snap + ':'):
                        plug_snap = plug.split(':')[0]
                        if plug_snap != '-' and \
                           plug_snap not in unseeded_snaps:
                            logging.debug("Snap %s is being used by %s. "
                                          "Switching it to stable track"
                                          % (snap, plug_snap))
                            snap_object['command'] = 'refresh'
                            snap_object['channel'] = 'stable'
                            break

                snap_list[snap] = snap_object

        return snap_list

    def update_snap_cb(self, client, change, _, user_data):
        index, count, progress_bar = user_data
        if not progress_bar:
            return

        # determine how much of this change has been done
        task_total = 0
        task_done = 0
        for task in change.get_tasks():
            task_total += task.get_progress_total()
            task_done += task.get_progress_done()

        task_fraction = task_done / task_total

        # determine how much total progress has been made
        total_fraction = (task_fraction / count) + (index / count)

        # change.get_tasks() can increase between callbacks so we must
        # avoid jumping backward in progress here
        if total_fraction > progress_bar.get_fraction():
            progress_bar.set_fraction(total_fraction)

    def update_snaps(self):
        # update status and progress bar
        def update_status(status):
            GLib.idle_add(self.label_details.set_label, status)

        def update_progress(progress_bar):
            progress_bar.pulse()
            return True

        update_status(_("Updating snaps"))

        progress_bar = None
        progress_timer = None

        progress_bars = self.progressbar_slot.get_children()
        if progress_bars and isinstance(progress_bars[0], Gtk.ProgressBar):
            progress_bar = progress_bars[0]
            progress_timer = GLib.timeout_add(100, update_progress,
                                              progress_bar)

        # populate snap_list with deb2snap transitions
        snap_list = self.get_snap_transitions()

        if progress_timer:
            GLib.source_remove(progress_timer)
            progress_bar.set_fraction(0)

        # (un)install (un)seeded snap(s)
        try:
            client = Snapd.Client()
            client.connect_sync()
            index = 0
            count = len(snap_list)
            for snap, snap_object in snap_list.items():
                command = snap_object['command']
                if command == 'refresh':
                    update_status(_("Refreshing %s snap" % snap))
                    client.refresh_sync(snap, snap_object['channel'],
                                        self.update_snap_cb,
                                        progress_callback_data=(index, count,
                                                                progress_bar))
                elif command == 'remove':
                    update_status(_("Removing %s snap" % snap))
                    client.remove_sync(snap, self.update_snap_cb,
                                       progress_callback_data=(index, count,
                                                               progress_bar))
                else:
                    update_status(_("Installing %s snap" % snap))
                    client.install_sync(snap, snap_object['channel'],
                                        self.update_snap_cb,
                                        progress_callback_data=(index, count,
                                                                progress_bar))
                index += 1
        except GLib.Error as e:
            logging.debug("error updating snaps (%s)" % e)
            GLib.idle_add(self.window_main.start_error, False,
                          _("Upgrade only partially completed."),
                          _("An error occurred while updating snaps. "
                            "Please check your network connection."))
            return

        # continue with the rest of the updates
        GLib.idle_add(self.window_main.start_available)

    def _action_done(self, action, authorized, success, error_string,
                     error_desc, trans_failed=False):

        # If the progress dialog should be closed automatically afterwards
        #settings = Gio.Settings.new("com.ubuntu.update-manager")
        #close_after_install = settings.get_boolean(
        #    "autoclose-install-window")
        # FIXME: confirm with mpt whether this should still be a setting
        #close_after_install = False

        if action == self.ACTION_INSTALL:
            if success and os.path.exists("/usr/bin/snap"):
                Thread(target=self.update_snaps).start()
            elif success:
                self.window_main.start_available()
            elif error_string:
                self.window_main.start_error(trans_failed, error_string,
                                             error_desc)
            else:
                # exit gracefuly, we can't just exit as this will trigger
                # a crash if system.exit() is called in a exception handler
                GLib.timeout_add(1, self.window_main.exit)
        else:
            if error_string:
                self.window_main.start_error(True, error_string, error_desc)
            elif success and os.path.exists("/usr/bin/snap"):
                self.window_main.duplicate_packages = self.get_deb2snap_dups()
                self.window_main.start_available()
            else:
                is_cancelled_update = not success
                self.window_main.start_available(is_cancelled_update)


# try aptdaemon
if os.path.exists("/usr/sbin/aptd") \
   and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ:
    # check if the gtkwidgets are installed as well
    try:
        from .InstallBackendAptdaemon import InstallBackendAptdaemon
    except ImportError:
        logging.exception("importing aptdaemon")
# try synaptic
if os.path.exists("/usr/sbin/synaptic") \
   and "UPDATE_MANAGER_FORCE_BACKEND_APTDAEMON" not in os.environ:
    try:
        from .InstallBackendSynaptic import InstallBackendSynaptic
    except ImportError:
        logging.exception("importing synaptic")


def get_backend(*args, **kwargs):
    """Select and return a package manager backend."""
    # try aptdaemon
    if (os.path.exists("/usr/sbin/aptd")
            and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ):
        # check if the gtkwidgets are installed as well
        try:
            return InstallBackendAptdaemon(*args, **kwargs)
        except NameError:
            logging.exception("using aptdaemon failed")
    # try synaptic
    if (os.path.exists("/usr/sbin/synaptic")
            and "UPDATE_MANAGER_FORCE_BACKEND_APTDAEMON" not in os.environ):
        try:
            return InstallBackendSynaptic(*args, **kwargs)
        except NameError:
            pass
    # nothing found, raise
    raise Exception("No working backend found, please try installing "
                    "aptdaemon or synaptic")



#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import datetime
from gi.repository import GLib
import http.client
from mock import Mock
import logging
import sys
import unittest
import yaml

from UpdateManager.Core.LivePatchSocket import LivePatchSocket, LivePatchFix


status0 = {'architecture': 'x86_64',
           'boot-time': datetime.datetime(2017, 6, 27, 11, 16),
           'client-version': '7.21',
           'cpu-model': 'Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz',
           'last-check': datetime.datetime(2017, 6, 28, 14, 23, 29, 683361),
           'machine-id': 123456789,
           'machine-token': 987654321,
           'uptime': '27h12m12s'}

status1 = {'architecture': 'x86_64',
           'boot-time': datetime.datetime(2017, 6, 27, 11, 16),
           'client-version': '7.21',
           'cpu-model': 'Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz',
           'last-check': datetime.datetime(2017, 6, 28, 14, 23, 29, 683361),
           'machine-id': 123456789,
           'machine-token': 987654321,
           'status': [{'kernel': '4.4.0-78.99-generic',
                       'livepatch': {'checkState': 'needs-check',
                                     'fixes': '',
                                     'patchState': 'nothing-to-apply',
                                     'version': '24.2'},
                       'running': True}],
           'uptime': '27h12m12s'}

status2 = {'architecture': 'x86_64',
           'boot-time': datetime.datetime(2017, 6, 27, 11, 16),
           'client-version': '7.21',
           'cpu-model': 'Intel(R) Core(TM) i7-6700HQ CPU @ 2.60GHz',
           'last-check': datetime.datetime(2017, 6, 28, 14, 23, 29, 683361),
           'machine-id': 123456789,
           'machine-token': 987654321,
           'status': [{'kernel': '4.4.0-78.99-generic',
                       'livepatch': {'checkState': 'checked',
                                     'fixes': '* CVE-2016-0001\n'
                                     '* CVE-2016-0002\n'
                                     '* CVE-2017-0001 (unpatched)\n'
                                     '* CVE-2017-0001',
                                     'patchState': 'applied',
                                     'version': '24.2'},
                       'running': True}],
           'uptime': '27h12m12s'}


class TestUtils(object):

    @staticmethod
    def __TimeoutCallback(user_data=None):
        user_data[0] = True
        return False

    @staticmethod
    def __ScheduleTimeout(timeout_reached, timeout_duration=10):
        return GLib.timeout_add(timeout_duration,
                                TestUtils.__TimeoutCallback,
                                timeout_reached)

    @staticmethod
    def WaitUntilMSec(instance, check_function, expected_result=True,
                      max_wait=500, error_msg=''):
        instance.assertIsNotNone(check_function)

        timeout_reached = [False]
        timeout_id = TestUtils.__ScheduleTimeout(timeout_reached, max_wait)

        result = None
        while not timeout_reached[0]:
            result = check_function()
            if result == expected_result:
                break
            GLib.MainContext.default().iteration(True)

        if result == expected_result:
            GLib.Source.remove(timeout_id)

        instance.assertEqual(expected_result, result, error_msg)


class MockResponse():

    def __init__(self, status, data):
        self.status = status
        self.data = data

    def read(self):
        return yaml.dump(self.data)


class TestLivePatchSocket(unittest.TestCase):

    def test_get_check_state(self):
        check_state = LivePatchSocket.get_check_state(status0)
        self.assertEqual(check_state, 'check-failed')
        check_state = LivePatchSocket.get_check_state(status1)
        self.assertEqual(check_state, 'needs-check')
        check_state = LivePatchSocket.get_check_state(status2)
        self.assertEqual(check_state, 'checked')

    def test_get_patch_state(self):
        patch_state = LivePatchSocket.get_patch_state(status0)
        self.assertEqual(patch_state, 'unknown')
        patch_state = LivePatchSocket.get_patch_state(status1)
        self.assertEqual(patch_state, 'nothing-to-apply')
        patch_state = LivePatchSocket.get_patch_state(status2)
        self.assertEqual(patch_state, 'applied')

    def test_get_fixes(self):
        fixes = LivePatchSocket.get_fixes(status0)
        self.assertEqual(fixes, [])
        fixes = LivePatchSocket.get_fixes(status1)
        self.assertEqual(fixes, [])
        fixes = LivePatchSocket.get_fixes(status2)
        self.assertEqual(fixes, [LivePatchFix('CVE-2016-0001'),
                                 LivePatchFix('CVE-2016-0002'),
                                 LivePatchFix('CVE-2017-0001 (unpatched)'),
                                 LivePatchFix('CVE-2017-0001')])

    def test_livepatch_fix(self):
        fix = LivePatchFix('CVE-2016-0001')
        self.assertEqual(fix.name, 'CVE-2016-0001')
        self.assertTrue(fix.patched)

        fix = LivePatchFix('CVE-2016-0001 (unpatched)')
        self.assertEqual(fix.name, 'CVE-2016-0001')
        self.assertFalse(fix.patched)

    def test_callback_not_active(self):
        mock_http_conn = Mock(spec=http.client.HTTPConnection)
        attrs = {'getresponse.return_value': MockResponse(400, None)}
        mock_http_conn.configure_mock(**attrs)

        cb_called = [False]

        def on_done(active, check_state, patch_state, fixes):
            cb_called[0] = True
            self.assertFalse(active)

        lp = LivePatchSocket(mock_http_conn)
        lp.get_status(on_done)

        mock_http_conn.request.assert_called_with(
            'GET', '/status?verbose=True')
        TestUtils.WaitUntilMSec(self, lambda: cb_called[0] is True)

    def test_callback_active(self):
        mock_http_conn = Mock(spec=http.client.HTTPConnection)
        attrs = {'getresponse.return_value': MockResponse(200, status2)}
        mock_http_conn.configure_mock(**attrs)

        cb_called = [False]

        def on_done(active, check_state, patch_state, fixes):
            cb_called[0] = True
            self.assertTrue(active)
            self.assertEqual(check_state, 'checked')
            self.assertEqual(patch_state, 'applied')
            self.assertEqual(fixes, [LivePatchFix('CVE-2016-0001'),
                                     LivePatchFix('CVE-2016-0002'),
                                     LivePatchFix('CVE-2017-0001 (unpatched)'),
                                     LivePatchFix('CVE-2017-0001')])

        lp = LivePatchSocket(mock_http_conn)
        lp.get_status(on_done)

        mock_http_conn.request.assert_called_with(
            'GET', '/status?verbose=True')
        TestUtils.WaitUntilMSec(self, lambda: cb_called[0] is True)


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == "-v":
        logging.basicConfig(level=logging.DEBUG)
    unittest.main()




import datetime
import gettext
gettext.install("update-manager")
from gettext import gettext as _


# the day on which the short support HWE stack goes EoL
HWE_EOL_DATE = datetime.date(2025, 4, 30)

# the day on which the next LTS first point release is available
# used to propose a release upgrade
NEXT_LTS_DOT1_DATE = datetime.date(2022, 7, 21)

# end of the month in which this LTS goes EoL
LTS_EOL_DATE = datetime.date(2025, 4, 30)


class Messages:
    UM_UPGRADE = _("""
There is a graphics stack installed on this system. An upgrade to a
configuration supported for the full lifetime of the LTS will become
available on %(date)s and can be installed by running 'update-manager'
in the Dash.
    """) % {'date': NEXT_LTS_DOT1_DATE.isoformat()}

    APT_UPGRADE = _("""
To upgrade to a supported (or longer-supported) configuration:

* Upgrade from Ubuntu 18.04 LTS to Ubuntu 20.04 LTS by running:
sudo do-release-upgrade %s

OR

* Switch to the current security-supported stack by running:
sudo apt-get install %s

and reboot your system.""")

    # this message is shown if there is no clear upgrade path via a
    # meta pkg that we recognize
    APT_SHOW_UNSUPPORTED = _("""
The following packages are no longer supported:
 %s

Please upgrade them to a supported HWE stack or remove them if you
no longer need them.
""")

    HWE_SUPPORTED = _("Your Hardware Enablement Stack (HWE) is "
                      "supported until %(month)s %(year)s.") % {
                          'month': LTS_EOL_DATE.strftime("%B"),
                          'year': LTS_EOL_DATE.year}

    HWE_SUPPORT_ENDS = _("""
Your current Hardware Enablement Stack (HWE) is going out of support
on %s.  After this date security updates for critical parts (kernel
and graphics stack) of your system will no longer be available.

For more information, please see:
http://wiki.ubuntu.com/2004_HWE_EOL
""") % HWE_EOL_DATE.isoformat()

    HWE_SUPPORT_HAS_ENDED = _("""
WARNING: Security updates for your current Hardware Enablement
Stack ended on %s:
 * http://wiki.ubuntu.com/2004_HWE_EOL
""") % HWE_EOL_DATE.isoformat()
